<?php

declare(strict_types=1);

namespace Drupal\decoupled_passkeys\Plugin\jsonrpc\Method;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysException;
use Drupal\decoupled_passkeys\Exception\DecoupledPasskeysUnexpectedValueException;
use Drupal\jsonrpc\Exception\JsonRpcException;
use Drupal\jsonrpc\MethodInterface;
use Drupal\jsonrpc\Object\Error;
use Drupal\jsonrpc\Object\ParameterBag;
use Drupal\jsonrpc\Plugin\JsonRpcMethodBase;
use Drupal\webauthn_framework\Services\Webauthn;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Webauthn\PublicKeyCredentialSource as PKCredentialSource;
use Webmozart\Assert\Assert;

/**
 * Authenticates via passkey.
 *
 * @JsonRpcMethod(
 *   id = "user.authenticate_request",
 *   usage = @Translation("Authenticates a user via passkey."),
 *   access = {"login by passkey"},
 *   params = {
 *     "request" = @JsonRpcParameterDefinition(
 *       schema={"type"="string"},
 *       required=true
 *     ),
 *     "userHandle" = @JsonRpcParameterDefinition(
 *        schema={"type"="string"},
 *        required=true *
 *     ),
 *   }
 * ),
 */
final class AuthenticateRequest extends JsonRpcMethodBase {

  /**
   * The current user account.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected AccountInterface $currentUser;

  /**
   * Logger factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $loggerFactory;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The Webauthn Framework service.
   *
   * @var \Drupal\webauthn_framework\Services\Webauthn
   */
  protected Webauthn $webauthn;

  /**
   * The request parameter.
   */
  private const PARAMETER_REQUEST = 'request';

  /**
   * The webauthn userHandle parameter.
   */
  private const PARAMETER_USER_HANDLE = 'userHandle';

  /**
   * Constructs a new EntityResourceBase object.
   *
   * @param array $configuration
   *   JSON-RPC config.
   * @param string $plugin_id
   *   The plugin ID.
   * @param \Drupal\jsonrpc\MethodInterface $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Core\Session\AccountInterface $account
   *   The current user.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The Watchdog logger.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\webauthn_framework\Services\Webauthn $webauthn
   *   The Webauthn Framework class.
   */
  public function __construct(array $configuration,
  string $plugin_id,
  MethodInterface $plugin_definition,
  AccountInterface $account,
  LoggerChannelFactoryInterface $logger_factory,
  EntityTypeManagerInterface $entity_type_manager,
  Webauthn $webauthn) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->currentUser = $account;
    $this->loggerFactory = $logger_factory->get('decoupled_passkeys');
    $this->entityTypeManager = $entity_type_manager;
    $this->webauthn = $webauthn;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): AuthenticateRequest {
    if (!$plugin_definition instanceof MethodInterface) {
      throw new DecoupledPasskeysUnexpectedValueException('Plugin definition should be MethodInterface!');
    }
    return new self(
      $configuration, $plugin_id, $plugin_definition,
      $container->get('current_user'),
      $container->get('logger.factory'),
      $container->get('entity_type.manager'),
      $container->get('webauthn_framework.webauthn')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function execute(ParameterBag $params) {
    try {

      // Ensure the parameters have been supplied.
      $authentication_request = $params->get(self::PARAMETER_REQUEST);
      Assert::stringNotEmpty($authentication_request, 'No value for request!');
      $user_handle = $params->get(self::PARAMETER_USER_HANDLE);
      Assert::stringNotEmpty($user_handle, 'No value for userHandle!');

      $pk_credential_source = $this->webauthn->authenticateRequest($user_handle, $authentication_request);

      if ($pk_credential_source instanceof PKCredentialSource) {
        $uuid = $pk_credential_source->userHandle;
        $user_storage = $this->entityTypeManager->getStorage('user');
        $account_array = $user_storage->loadByProperties(['uuid' => $uuid]);
        $account = reset($account_array);
        Assert::isInstanceOf($account, '\Drupal\user\UserInterface');
        /** @var \Drupal\user\UserInterface $account */

        user_login_finalize($account);
        return json_encode($account);
      }
      throw new DecoupledPasskeysException('Failed to get PKCredentialSource!');
    }
    /* @phpstan-ignore-next-line Thrown exceptions in catch block must bundle previous exception. */
    catch (\LogicException $e) {
      $error_message = $e->getMessage();
      $this->loggerFactory->error($error_message);
      $error = Error::internalError();
      throw JsonRpcException::fromError($error);
    }
  }

  /**
   * {@inheritdoc}
   */
  public static function outputSchema(): array {
    return ['type' => 'string'];
  }

}
